--	Destination
--	Represents a potential destination for addon output.


local major, minor, libTable = LibStub("LibBuilder-1.0"):GetLib()
if not libTable.canLoad then return end


local Destination, destination = libTable.classes:NewClass("Destination")


Destination.nameToFrame = {}		--	map of tab names to frame tables


destination.type = nil
destination.value = nil


---	Creates a new Destination object.
--	@see	SetValue
--	@param	value		The value for the new destination. See SetValue
--	@return instance	The new object.
function Destination:New(value)
	local instance = {}
	setmetatable(instance, self.instanceMetatable)
	
	if value then
		instance:SetValue(value)
	end
	
	instance:SetExclusive(true)
	
	return instance
end


--	Internal; refreshes the mapping of tab names to frame objects
--	For now we're just going to call this each time a Print is requested
function Destination:RefreshTabMap()
	for i=1,NUM_CHAT_WINDOWS do
		local name, _, _, _, _, _, isShown, _, isDocked = GetChatWindowInfo(i)
		
		--	NOTE: See FCF_OpenNewWindow
		
		--	Skip windows that aren't visible since they 
		--	aren't condidered "available" anyway.
		--	Just remember that we have to watch when those attributes change.
		if (isShown or isDocked) then
			self.nameToFrame[name] = _G["ChatFrame"..i]
		end
	end
end


--	Internal; retrieves the frame associated with a named tab.
--	@param	name	Name of the tab associated with the frame.
--	@param	frame	The frame.
function Destination:GetFrame(name)
	--	TODO: hook something so we don't have to do this every time
	self:RefreshTabMap()
	
	return self.nameToFrame[name]
end


local checks = {
	tab =		function(v) return type(v)=="string" and #v>0 end, 
	frame =		function(v) local t=type(v); return (t=="table" and v.AddMessage) or (t=="number" and v==floor(v) and 0<=v and v<=NUM_CHAT_WINDOWS) end, 
	callback =	function(v) return type(v)=="function" end,
}


--	Internal; checks if the tentative value is valid.
--	@param	value	The tentative value to test.
--	@param	dType	The destination type to use for the test. See SetType
--	@see	SetType
--	@return	isValid	True if the specified value is valid, false otherwise
function Destination:IsValidValue(value, dType)
	if dType then
		return checks[dType](value)
	else
		for _, check in pairs(checks) do
			if check(value) then
				return true
			end
		end
		
		--	none of the checks returned true
		return false
	end
end


---	Sets the type for this destination.
--	Valid types:
--	nil					Generic, will guess type based on value.
--	"tab"				Accepts the name of a tab; messages will print to the associated frame.
--	"frame"				Accepts a frame table or index; messages will print to that frame.
--	"callback"			Accepts a function closure; it will be called with the message as an argument.
--	@param	newType		The new type.
function destination:SetType(newType)
	if newType == "nil" then newType = nil end
	
	assert(newType == nil or checks[newType] ~= nil, "Invalid Destination type")
	
	self.value = nil
	self.type = newType
end


---	Returns this Destination's type.
--	@see	SetType
--	@return	dType	Type of this destination.
function destination:GetType()
	--	if type is nil, we'll have to check the value.
	if self.type ~= nil then
		return self.type
	end
	
	local t = type(self.value)
	
	if t == "string" then
		return "tab"
	elseif t == "function" then
		return "callback"
	elseif t == "table" or t == "number" then
		return "frame"
	end
end


---	Sets the value for this Destination.
--	If this Destination's type is "tab", the value must be a non-empty string.
--	If the type is "frame", the value must be the frame object, or a number 
--	between 0 and NUM_CHAT_WINDOWS (inclusive) where 0 indicates DEFAULT_CHAT_FRAME, 
--	and other numbers indicate the frame ending with that value (eg. 3 -> ChatFrame3)
--	If the type is "callback", the value must be a function closure designed to take 
--	arguments of the form (separator, defaultColorString, message); except that 
--	if the 'raw' flag of this destination is set, 'message' will be replaced by a 
--	variable length list of message segments.
--	@param value	The new value.
function destination:SetValue(value)
	assert(Destination:IsValidValue(value, self.type), "Invalid Destination value")

	self.value = value
end


---	Checks if this Destination is currently available.
--	@return	isAvailable		True if the destination is available, false otherwise.
function destination:IsAvailable()
	local vt = type(self.value)
	
	if vt == nil then
		return false
	end
	
	local dt = self:GetType()
	
	if dt == "tab" then
		return Destination:GetFrame(self.value) and true or false
	elseif dt == "callback" then
		return vt == "function"
	elseif dt == "frame" then
		return self.value and true or false
	end
end


---	Changes the 'exclusive' flag for this destination. Once 
--	Stream finds an exclusive destination, it will stop looking for more. 
--	Note that other, non-exclusive destinations will still recieve output 
--	if they are located before an exclusive destination in the priority list.
--	@param	isExclusive		True if this destination should be exclusive, false otherwise.
function destination:SetExclusive(isExclusive)
	assert(type(isExclusive) == "boolean")
	self.isExclusive = isExclusive
end


---	Checks if this is an exclusive Destination.
--	return isExclusive		True if this Destination is exclusive, false otherwise
function destination:IsExclusive()
	return self.isExclusive
end


---	Changes the 'raw' flag for this Destination.
--	Raw destinations will recieve formatting information plus the raw arugments 
--	passed to stream:Print, instead of receiving a pre-formmated string.
--	@param	isRaw		True if this Destination is raw, false otherwise.
function destination:SetRaw(isRaw)
	assert(type(isRaw) == "boolean")
	self.isRaw = isRaw
end


---	Checks if this destination is raw.
--	@see	SetRaw
--	@return isRaw		True if this destination is raw, false otherwise.
function destination:IsRaw()
	return self.isRaw or false
end


---	Sends print data to this Destination.
--	This does not check if the Destination is raw; the 
--	data must already be passed with that taken into account
--	@param	...		The data to send
function destination:Send(...)
	local t = self:GetType()
	
	if t == "tab" then
		Destination:GetFrame(self.value):AddMessage(...)
	elseif t == "callback" then
		self.value(...)
	elseif t == "frame" then
		local frame = self.value
		
		if type(frame) == "number" then
			if frame == 0 then
				frame = DEFAULT_CHAT_FRAME
			else
				frame = _G["ChatFrame"..frame]
			end
		end
		
		frame:AddMessage(...)
	end
end
