

local major, minor, libTable = LibStub("LibBuilder-1.0"):GetLib()
if not libTable.canLoad then return end


local debug = libTable.debug

local Stream, stream = libTable.classes:NewClass("Stream")


function Stream:New()
	local instance = {}
	setmetatable(instance, self.instanceMetatable)
	instance.destinations = {}
	return instance
end


---	Returns a list of Destinations to which 
--	messages should currently be sent.
--	@return targets		The list of target destinations, to 
--						which output should currently be sent.
function stream:PickDestinations()
	local active = {}
	
	for i, destination in ipairs(self.destinations) do
		if destination:IsAvailable() then
			table.insert(active, destination)
		
			--	If it's exclusive, don't look for more
			if destination:IsExclusive() then
				break
			end
		end
	end
	
	--debug("PickDestinations returning", active, #active, "elements")
	
	return active
end


---	Sends the provided items to the appropriate Destinations, 
--	with the formatting specified by each Destination.
--	If no arguments are provided, no Destinations will be triggered.
--	@param	...		A variable length list of items to print.
function stream:Print(...)
	--debug("Print:", ...)

	if select("#", ...) == 0 then
		--	just ignore the call (don't even display prefix)
		return
	end
	
	local separator = self.separator
	local prefix = self.prefix
	local colorString = self.colorString

	--	don't build the message yet, in case there are no active 
	--	destinations or they are all raw
	local message
	
	local destinations = self:PickDestinations()
	
	for i, destination in ipairs(destinations) do
		if destination:IsRaw() then
			if destination:GetType() == "function" then
				destination:Send(separator, colorString, prefix, ...)
			else
				destination:Send(...)
			end
		else
			--	build the message if we haven't already
			if not message then
				message = self:BuildMessage(separator, colorString, prefix, ...)
			end
			
			--debug("Sending destination", destination, "message", message)
			
			destination:Send(message)
		end
	end
end


---	Builds a message from the given segments.
--	This should only be used when a default color is not set.
--	Also, this function will not reset colors after separators.
--	@param	segments...		The segments to put together.
function stream:BuildMessage(separator, colorString, prefix, ...)
	--debug("BuildMessage, units:", separator, colorString, prefix, ...)

	if colorString then
		return libTable.classes.Painter:Paint(separator, colorString, prefix, ...)
	end

	--	the vararg might end with nils, so count them before packing
	local unitCount = select("#", ...)
	
	--	we already checked for an empty arglist

	local units
	
	if prefix then
		units = { prefix, ... }
		unitCount = unitCount + 1
	else
		units = { ... }
	end
	
	for i, unit in pairs(units) do
		local t = type(unit)
		
		if t == "string" or t == "number" then
			--	leave existing unit
		else
			units[i] = tostring(unit) or "nil"
		end
	end
	
	return table.concat(units, separator or " ")
end


---	Sets the prefix for messages sent through this stream.
--	Pass nil to remove the prefix.
--	If a value is provided for any color argument, the 
--	others will default to 0.
--	@param	prefix		The new prefix; an optional colorString can be embedded.
--	@param	r			Optional; Red color component for the prefix.
--	@param	g			Optional; Green color component for the prefix.
--	@param	b			Optional; Blue color component for the prefix.
function stream:SetPrefix(prefix, r, g, b)
	assert(prefix == nil or type(prefix) == "string" and #prefix > 0)
	
	if r or g or b then
		assert(prefix)
		prefix = string.format("|cff%02x%02x%02x%s|r", r or 0, g or 0, b or 0, prefix)
	end

	self.prefix = prefix
end


---	Sets the default color for messages sent through this stream.
--	All components default to 0.
--	@param	r			Optional; Red color component for the prefix.
--	@param	g			Optional; Green color component for the prefix.
--	@param	b			Optional; Blue color component for the prefix.
function stream:SetColor(r, g, b)
	self.colorString = string.format("ff%02x%02x%02x", r or 0, g or 0, b or 0)
end


---	Sets the color string for messages sent through this stream.
--	Alternative to SetColor.  Color strings should include the alpha 
--	value but not the UI escape sequences, eg. "ffffff00".
--	@param	colorString		The new colorString.
function stream:SetColorString(colorString)
	if colorString == nil then
		self.colorString = nil
		return
	end
	
	assert(type(colorString) == "string")
	
	if #colorString == 6 then
		colorString = "ff"..colorString
	end
	
	assert(colorString:match("%x%x%x%x%x%x%x%x"))

	self.colorString = colorString
end


---	Removes all Destinations from this Stream.
function stream:ClearDestinations()
	self.destinations = {}
end


---	Removes a Destination from this Stream.
--	@param		target	Position or object to remove.
function stream:RemoveDestination(target)
	assert(target)
	
	local len = #self.destinations

	local index
	
	--	find and remove the Destination
	if type(target) == "number" then
		self.destinations[target] = nil
		index = target
	else
		--	we have to search for it
		index = 1

		while index <= len do
			local test = self.destinations[index]
			
			if test == target then
				self.destinations[index] = nil
				break
			end
			
			index = index + 1
		end
	end
	
	--	now shift everything down
	while index <= len do
		self.destinations[index] = self.destinations[index+1]
		index = index + 1
	end
end


---	Adds a Destination to the Stream.
--	@param	new			The new Destination.
--	@param	position	Optional; the position to add the new Destination. Defaults to the end.
function stream:AddDestination(new, position)
	--	TODO: make sure it's not a duplicate
	
	local len = #self.destinations

	if position then
		assert(0 < position and position <= len + 1)
	else
		position = len + 1
	end
	
	for i = len, position, -1 do
		self.destinations[i+1] = self.destinations[i]
	end
	
	self.destinations[position] = new
end
